import os
from functools import lru_cache

from pydantic import BaseModel

from monarch_py.implementations.solr.solr_implementation import SolrImplementation
from monarch_py.implementations.spacy.spacy_implementation import SpacyImplementation
from monarch_py.service.semsim_service import SemsimianService


class Settings(BaseModel):
    solr_host: str = os.getenv("SOLR_HOST") if os.getenv("SOLR_HOST") else "127.0.0.1"
    solr_port: str = os.getenv("SOLR_PORT") if os.getenv("SOLR_PORT") else 8983
    solr_url: str = os.getenv("SOLR_URL") if os.getenv("SOLR_URL") else f"http://{solr_host}:{solr_port}/solr"
    phenio_db_path: str = os.getenv("PHENIO_DB_PATH") if os.getenv("PHENIO_DB_PATH") else "/data/phenio.db"

    semsim_server_host: str = os.getenv("SEMSIM_SERVER_HOST", "127.0.0.1")
    semsim_server_port: str = os.getenv("SEMSIM_SERVER_PORT", 9999)


settings = Settings()


@lru_cache(maxsize=1)
def solr():
    return SolrImplementation(settings.solr_url)


@lru_cache(maxsize=1)
def semsimian():
    return SemsimianService(
        semsim_server_host=settings.semsim_server_host,
        semsim_server_port=settings.semsim_server_port,
        entity_implementation=solr(),
    )


@lru_cache(maxsize=1)
def oak():
    return NotImplementedError("OAK is temporarily disabled")


#    oak_implementation = OakImplementation()
#    oak_implementation.init_phenio_adapter(force_update=False, phenio_path=settings.phenio_db_path)
#    return oak_implementation


@lru_cache(maxsize=1)
def spacyner():
    spacy_implementation = SpacyImplementation()
    spacy_implementation.init_spacy(grounding_implementation=solr())
    return spacy_implementation



from fastapi import APIRouter, Path, Query

from monarch_py.api.additional_models import (
    SemsimCompareRequest,
    SemsimMetric,
    SemsimSearchRequest,
    SemsimSearchGroup,
    SemsimMultiCompareRequest,
    SemsimDirectionality,
)
from monarch_py.api.config import semsimian, solr
from monarch_py.api.utils.similarity_utils import parse_similarity_prefix
from monarch_py.datamodels.category_enums import AssociationPredicate, EntityCategory
from monarch_py.datamodels.model import SearchResults

router = APIRouter(tags=["semsim"], responses={404: {"description": "Not Found"}})


@router.get("/autocomplete")
def autocomplete(
    q: str = Query(
        default="*:*",
        title="Query string to autocomplete against",
        examples=["fanc", "ehler"],
    ),
) -> SearchResults:
    """
    Autocomplete for semantic similarity lookups, prioritizes entities which have direct phenotype associations.
    Note: This API endpoint is experimental and may evolve or disappear over time.

    Args:
        q (str): Query string to autocomplete against

    Returns:
        SearchResults
    """
    response = solr().autocomplete(
        q=q,
        category=[EntityCategory.DISEASE, EntityCategory.GENE, EntityCategory.PHENOTYPIC_FEATURE],
        prioritized_predicates=[AssociationPredicate.HAS_PHENOTYPE],
    )
    return response


@router.get("/compare/{subjects}/{objects}")
def _compare(
    subjects: str = Path(..., title="List of subjects for comparison"),
    objects: str = Path(..., title="List of objects for comparison"),
    metric: SemsimMetric = Query(SemsimMetric.ANCESTOR_INFORMATION_CONTENT, title="Similarity metric to use"),
):
    """Get pairwise similarity between two sets of terms

    <b>Args:</b> <br>
        subjects (str, optional): List of subjects for comparison. Defaults to "". <br>
        objects (str, optional): List of objects for comparison. Defaults to "". <br>
        metric (str, optional): Similarity metric to use. Defaults to "ancestor_information_content".

    <b>Returns:</b> <br>
        TermSetPairwiseSimilarity: Pairwise similarity between subjects and objects
    """
    print(
        f"""
    Running semsim compare:
        subjects: {subjects.split(',')}
        objects: {objects.split(',')}
        metric: {metric}
    """
    )
    results = semsimian().compare(
        subjects=subjects.split(","),
        objects=objects.split(","),
        metric=metric,
    )
    return results


@router.post("/compare")
def _post_compare(request: SemsimCompareRequest):
    """
        Pairwise similarity between two sets of terms <br>
        <br>
        Example: <br>
    <pre>
    {
        "subjects": ["MP:0010771", "MP:0002169"],
        "objects": ["HP:0004325"]
        "metric": "ancestor_information_content"
    }
    </pre>
    """
    return semsimian().compare(subjects=request.subjects, objects=request.objects, metric=request.metric)


# add a multicompare post endpoint
@router.post("/multicompare")
def _post_multicompare(request: SemsimMultiCompareRequest):
    """
        Pairwise similarity between two sets of terms <br>
        <br>
        Example: <br>
    <pre>
    {
        "metric": "ancestor_information_content",
        "subjects": ["HP:0002616","HP:0001763","HP:0004944","HP:0010749","HP:0001533","HP:0002020","HP:0012450"],
        "object_sets": [
            {
            "id": "MGI:2441732",
            "label": "Adgrg7",
            "phenotypes": ["MP:0011965","MP:0002834","MP:0003731","MP:0011962","MP:0011960","MP:0008489","MP:0003291","MP:0001262"]
            },
            {
            "id": "MGI:87909",
            "label": "Acta2",
            "phenotypes": ["MP:0002834","MP:0003070","MP:0004022","MP:0004021","MP:0003026","MP:0006264","MP:0000230","MP:0000233","MP:0000272","MP:0009862"]
            }
        ]
    }
    </pre>
    """
    return semsimian().multi_compare(request)


@router.get("/search/{termset}/{group}")
def _search(
    termset: str = Path(..., title="Termset to search"),
    group: SemsimSearchGroup = Path(..., title="Group of entities to search within (e.g. Human Genes)"),
    metric: SemsimMetric = Query(SemsimMetric.ANCESTOR_INFORMATION_CONTENT, title="Similarity metric to use"),
    directionality: SemsimDirectionality = Query(
        SemsimDirectionality.BIDIRECTIONAL, title="Directionality of the search"
    ),
    limit: int = Query(default=10, ge=1, le=50),
):
    """Search for terms in a termset

    <b>Args:</b> <br>
        termset (str): Comma separated list of term IDs to find matches for. <br>
        group (str): Group of entities to search within (e.g. Human Genes) <br>
        metric: (str, optional): Similarity metric to use. Defaults to "ancestor_information_content". <br>
        limit (int, optional): Limit the number of results. Defaults to 10.

    <b>Returns:</b> <br>
        List[str]: List of matching terms
    """
    terms = [term.strip() for term in termset.split(",")]
    results = semsimian().search(
        termset=terms, prefix=parse_similarity_prefix(group), metric=metric, directionality=directionality, limit=limit
    )
    return results


@router.post("/search")
def _post_search(request: SemsimSearchRequest):
    """
        Search for terms in a termset <br>
        <br>
        Example: <br>
    <pre>
    {
      "termset": ["HP:0002104", "HP:0012378", "HP:0012378", "HP:0012378"],
      "group": "Human Diseases",
      "metric": "ancestor_information_content",
      "limit": 5
    }
    </pre>
    """
    return semsimian().search(
        termset=request.termset,
        prefix=parse_similarity_prefix(request.group.value),
        metric=request.metric,
        limit=request.limit,
    )



from typing import Union

from fastapi import APIRouter, HTTPException, Path, Query, Response

from monarch_py.api.config import solr
from monarch_py.api.additional_models import OutputFormat
from monarch_py.datamodels.model import HistoPheno
from monarch_py.utils.format_utils import to_tsv

router = APIRouter(
    tags=["histopheno"],
    responses={404: {"description": "Not Found"}},
)


@router.get("/{id}")
async def _get_histopheno(
    id: str = Path(
        title="ID of the entity to get histopheno data for",
        examples=["MONDO:0019391"],
    ),
    format: OutputFormat = Query(
        default=OutputFormat.json,
        title="Output format for the response",
        examples=["json", "tsv"],
    ),
) -> Union[HistoPheno, str]:
    """Retrieves the entity with the specified id"""
    response = solr().get_histopheno(id)
    if response is None:
        raise HTTPException(status_code=404, detail="Entity not found")
    if format == OutputFormat.json:
        return response
    elif format == OutputFormat.tsv:
        tsv = ""
        for row in to_tsv(response, print_output=False):
            tsv += row
        return Response(content=tsv, media_type="text/tab-separated-values")



from typing import List, Optional
from typing_extensions import Annotated

import pystow
import typer

from monarch_py.datamodels.category_enums import (
    AssociationCategory,
    AssociationPredicate,
    EntityCategory,
    MappingPredicate,
)
from monarch_py.utils.solr_cli_utils import (
    ensure_solr,
    get_solr,
    solr_status,
    start_solr,
    stop_solr,
)
from monarch_py.utils.utils import console, set_log_level
from monarch_py.utils.format_utils import format_output
from monarch_py.utils import cli_fields as fields

solr_app = typer.Typer()
monarchstow = pystow.module("monarch")


@solr_app.callback(invoke_without_command=True)
def callback(
    ctx: typer.Context,
    quiet: Annotated[bool, typer.Option("--quiet", "-q", help="Set log level to warning")] = False,
    debug: Annotated[bool, typer.Option("--debug", "-d", help="Set log level to debug")] = False,
):
    if ctx.invoked_subcommand is None:
        typer.secho(
            f"\n\tNo command specified\n\tTry `monarch solr --help` for more information.\n",
            fg=typer.colors.YELLOW,
        )
        raise typer.Exit()
    log_level = "DEBUG" if debug else "WARNING" if quiet else "INFO"
    set_log_level(log_level)


############################
### SOLR DOCKER COMMANDS ###
############################


@solr_app.command("start")
def start(update: bool = False):
    """Starts a local Solr container."""
    ensure_solr(overwrite=False)
    start_solr()


@solr_app.command("stop")
def stop():
    """Stops the local Solr container."""
    stop_solr()
    raise typer.Exit()


@solr_app.command("status")
def status():
    """Checks the status of the local Solr container."""
    solr_status()
    raise typer.Exit()


@solr_app.command("download")
def download(
    version: Annotated[
        str,
        typer.Argument(
            help="The version of the Solr KG to download (latest, dev, or a specific version)",
        ),
    ] = "latest",
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite",
            help="Overwrite the existing Solr KG if it exists",
        ),
    ] = False,
):
    """Download the Monarch Solr KG."""
    ensure_solr(version, overwrite)
    raise typer.Exit()


###########################
### SOLR QUERY COMMANDS ###
###########################


@solr_app.command("entity")
def entity(
    entity_id: Annotated[
        str,
        typer.Argument(help="The identifier of the entity to be retrieved"),
    ],
    extra: Annotated[
        bool,
        typer.Option(
            "--extra",
            "-e",
            help="Include extra fields in the output (association_counts and node_hierarchy)",
        ),
    ] = False,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Retrieve an entity by ID
    """

    solr = get_solr(update=False)
    response = solr.get_entity(entity_id, extra)
    if response is not None:
        format_output(fmt, response, output)
    else:
        console.print(f"\n[bold red]Entity {id} not found.[/]\n")
        raise typer.Exit(1)


@solr_app.command("associations")
def associations(
    category: Annotated[
        Optional[List[AssociationCategory]],
        typer.Option(
            "--category",
            "-c",
            help="Category to get associations for",
        ),
    ] = None,
    subject: Annotated[
        Optional[List[str]],
        typer.Option(
            "--subject",
            "-s",
            help="Subject ID to get associations for",
        ),
    ] = None,
    predicate: Annotated[
        Optional[List[AssociationPredicate]],
        typer.Option(
            "--predicate",
            "-p",
            help="Predicate ID to get associations for",
        ),
    ] = None,
    object: Annotated[
        Optional[List[str]],
        typer.Option(
            "--object",
            "-o",
            help="Object ID to get associations for",
        ),
    ] = None,
    entity: Annotated[
        Optional[List[str]],
        typer.Option(
            "--entity",
            "-e",
            help="Entity (subject or object) ID to get associations for",
        ),
    ] = None,
    direct: Annotated[
        bool,
        typer.Option(
            "--direct",
            "-d",
            help="Whether to exclude associations with subject/object as ancestors",
        ),
    ] = False,
    compact: Annotated[
        bool,
        typer.Option(
            "--compact",
            "-C",
            help="Whether to return a compact representation of the associations",
        ),
    ] = False,
    limit: fields.LimitOption = 20,
    offset: fields.OffsetOption = 0,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Paginate through associations
    """
    kwargs = locals()
    kwargs.pop("fmt", None)
    kwargs.pop("output", None)

    solr = get_solr(update=False)
    response = solr.get_associations(**kwargs)
    format_output(fmt, response, output)


@solr_app.command("multi-entity-associations")
def multi_entity_associations(
    entity: Annotated[
        Optional[List[str]],
        typer.Option(
            "--entity",
            "-e",
            help="Comma-separated list of entities",
        ),
    ] = None,
    counterpart_category: Annotated[
        Optional[List[str]],
        typer.Option("--counterpart-category", "-c", help="A comma-separated list of counterpart categories"),
    ] = None,
    limit: fields.LimitOption = 20,
    offset: fields.OffsetOption = 0,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Paginate through associations for multiple entities
    """
    kwargs = locals()
    kwargs.pop("fmt", None)
    kwargs.pop("output", None)
    kwargs["limit_per_group"] = kwargs.pop("limit")

    solr = get_solr(update=False)
    response = solr.get_multi_entity_associations(**kwargs)
    format_output(fmt, response, output)


@solr_app.command("search")
def search(
    q: fields.QueryOption = ":*",
    category: Annotated[
        Optional[List[EntityCategory]],
        typer.Option("--category", "-c"),
    ] = None,
    in_taxon_label: Annotated[
        Optional[str],
        typer.Option("--in-taxon-label", "-t"),
    ] = None,
    facet_fields: Annotated[
        Optional[List[str]],
        typer.Option("--facet-fields", "-ff"),
    ] = None,
    facet_queries: Annotated[
        Optional[List[str]],
        typer.Option("--facet-queries", "-fq"),
    ] = None,
    limit: fields.LimitOption = 20,
    offset: fields.OffsetOption = 0,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
    # sort: str = typer.Option(None, "--sort", "-s"),
):
    """
    Search for entities
    """
    kwargs = locals()
    kwargs.pop("fmt", None)
    kwargs.pop("output", None)

    solr = get_solr(update=False)
    response = solr.search(**kwargs)
    format_output(fmt, response, output)


@solr_app.command("autocomplete")
def autocomplete(
    q: Annotated[str, typer.Argument(help="Query string to autocomplete against")],
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Return entity autcomplete matches for a query string
    """
    solr = get_solr(update=False)
    response = solr.autocomplete(q)
    format_output(fmt, response, output)


@solr_app.command("histopheno")
def histopheno(
    subject: Annotated[str, typer.Argument(help="The subject of the association")],
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Retrieve the histopheno associations for a given subject
    """

    if not subject:
        console.print("\n[bold red]Subject ID required.[/]\n")
        raise typer.Exit(1)

    solr = get_solr(update=False)
    response = solr.get_histopheno(subject)
    format_output(fmt, response, output)


@solr_app.command("association-counts")
def association_counts(
    entity_id: Annotated[str, typer.Argument(help="The entity to get association counts for")],
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Retrieve the association counts for a given entity
    """
    solr = get_solr(update=False)
    response = solr.get_association_counts(entity_id)
    format_output(fmt, response, output)


@solr_app.command("association-table")
def association_table(
    entity_id: Annotated[
        str,
        typer.Argument(
            help="The entity to get associations for",
        ),
    ],
    category: Annotated[
        AssociationCategory,
        typer.Argument(
            help="The association category to get associations for, ex. biolink:GeneToPhenotypicFeatureAssociation",
        ),
    ],
    q: Annotated[
        Optional[str],
        typer.Option("--query", "-q"),
    ] = None,
    traverse_orthologs: Annotated[
        bool,
        typer.Option(
            "--traverse-orthologs",
            "-t",
            help="Whether to traverse orthologs when getting associations",
        ),
    ] = False,
    sort: Annotated[
        Optional[List[str]],
        typer.Option("--sort", "-s"),
    ] = None,
    limit: fields.LimitOption = 20,
    offset: fields.OffsetOption = 0,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    solr = get_solr(update=False)
    response = solr.get_association_table(
        entity=entity_id,
        category=category,
        traverse_orthologs=traverse_orthologs,
        sort=sort,
        q=q,
        limit=limit,
        offset=offset,
    )
    format_output(fmt, response, output)


@solr_app.command("mappings")
def mappings(
    entity_id: Annotated[
        Optional[List[str]],
        typer.Option("--entity-id", "-e", help="entity ID to get mappings for"),
    ] = None,
    subject_id: Annotated[
        Optional[List[str]],
        typer.Option("--subject-id", "-s", help="subject ID to get mappings for"),
    ] = None,
    predicate_id: Annotated[
        Optional[List[MappingPredicate]],
        typer.Option("--predicate-id", "-p", help="predicate ID to get mappings for"),
    ] = None,
    object_id: Annotated[
        Optional[List[str]],
        typer.Option("--object-id", "-o", help="object ID to get mappings for"),
    ] = None,
    mapping_justification: Annotated[
        Optional[List[str]],
        typer.Option(
            "--mapping-justification",
            "-m",
            help="mapping justification to get mappings for",
        ),
    ] = None,
    limit: fields.LimitOption = 20,
    offset: fields.OffsetOption = 0,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    args = locals()
    args.pop("fmt", None)
    args.pop("output", None)

    solr = get_solr(update=False)
    response = solr.get_mappings(**args)
    format_output(fmt, response, output)



import re
from monarch_py.datamodels.model import ExpandedCurie
from monarch_py.service.curie_service import converter

__all__ = [
    "get_expanded_curie",
]


def expansion_patch(url: str):
    """Any short term patches that we need while waiting for upstream fixes"""

    # This fixes MGI urls that take the form of
    # https://identifiers.org/MGI/97486
    # rather than https://identifiers.org/MGI:97486
    pattern = r"/MGI/(\d+)$"
    replacement = r"/MGI:\1"
    return re.sub(pattern, replacement, url)


def get_uri(id: str) -> str:
    """Returns the URI for the given CURIE."""
    return expansion_patch(converter.expand(id))


def get_expanded_curie(id: str) -> ExpandedCurie:
    """Returns the URI for the given CURIE."""
    return ExpandedCurie(id=id, url=converter.expand(id))



from abc import ABC
from typing import List

from monarch_py.datamodels.model import MappingResults


class MappingInterface(ABC):
    def get_mappings(
        self,
        entity_id: List[str] = None,
        subject_id: List[str] = None,
        predicate_id: List[str] = None,
        object_id: List[str] = None,
        mapping_justification: List[str] = None,
    ) -> MappingResults:
        """
        Get SSSOM Mappings based on the provided constraints

        Args:
            entity_id: Filter to only mappings matching the specified entity IDs. Defaults to None.
            subject_id: Filter to only mappings matching the specified subject IDs. Defaults to None.
            predicate_id: Filter to only mappings matching the specified predicate IDs. Defaults to None.
            object_id: Filter to only mappings matching the specified object IDs. Defaults to None.
            mapping_justification: Filter to only mappings matching the specified mapping justifications. Defaults to None.
        """
        raise NotImplementedError



import pkgutil
import re
from typing import List

import yaml
from monarch_py.datamodels.model import AssociationTypeMapping
from pydantic import TypeAdapter


class AssociationTypeMappings:
    __instance = None

    def __init__(self):
        if AssociationTypeMappings.__instance is not None:
            raise Exception("AssociationTypeMappings is a singleton class, use getInstance() to get the instance.")
        else:
            AssociationTypeMappings.__instance = self
            self.mappings = None
            self.load_mappings()

    @staticmethod
    def get_mappings():
        if AssociationTypeMappings.__instance is None:
            AssociationTypeMappings()
        return AssociationTypeMappings.__instance.mappings

    def get_mapping(self, category: str):
        if AssociationTypeMappings.__instance is None:
            AssociationTypeMappings()
        for mapping in self.mappings:
            if mapping.category == category:
                return mapping

    def load_mappings(self):
        mapping_data = pkgutil.get_data(__package__, "./association_type_mappings.yaml")
        mapping_data = yaml.load(mapping_data, Loader=yaml.FullLoader)
        adapter = TypeAdapter(List[AssociationTypeMapping])
        self.mappings = adapter.validate_python(mapping_data)


def get_association_type_mapping_by_query_string(
    query_string: str,
) -> AssociationTypeMapping:
    """
    Get the association type mapping for a given query string, splitting the category and predicate components apart
    Args:
        query_string: A solr query string to parse apart for category and predicate

    Returns: An AssociationTypeMapping instance appropriate for the given query string
    Raises: ValueError if no match is found
    """

    category = parse_query_string_for_category(query_string)

    matching_types = [a_type for a_type in AssociationTypeMappings.get_mappings() if a_type.category == category]

    if len(matching_types) == 0:
        raise ValueError(f"No matching association type found for query string: [{query_string}]")
    elif len(matching_types) > 1:
        raise ValueError(f"Too many association types found for query string: [{query_string}]")
    else:
        return matching_types[0]


def get_solr_query_fragment(agm: AssociationTypeMapping) -> str:
    return f'category:"{agm.category}"'


def get_sql_query_fragment(agm: AssociationTypeMapping) -> str:
    return f'category = "{agm.category}"'


def parse_query_string_for_category(
    query_string: str,
) -> str:
    categories = []

    pattern = re.compile(r'(category):\s*"?([\w:]+)"?')
    for match in re.findall(pattern, query_string):
        if match[0] == "category":
            categories.append(match[1])

    # Check if both categories and predicates were found
    if not categories:
        raise ValueError("No categories or predicates found in query string")

    if len(categories) > 1:
        raise ValueError(f"Multiple categories found in query string: {query_string}")

    return categories[0]



from enum import Enum
from typing import Annotated, Optional

import typer

# Common options

QueryOption = Annotated[
    str,
    typer.Option(
        "--query",
        "-q",
    ),
]

LimitOption = Annotated[
    int,
    typer.Option(
        "--limit",
        "-l",
        help="The number of results to return",
    ),
]

OffsetOption = Annotated[
    int,
    typer.Option(
        "--offset",
        help="The offset of the first result to be retrieved",
    ),
]


class OutputFormat(str, Enum):
    json = "json"
    yaml = "yaml"
    tsv = "tsv"
    table = "table"


FormatOption = Annotated[
    OutputFormat,
    typer.Option(
        "--format",
        "-f",
        help="The format of the output (json, yaml, tsv, table)",
    ),
]

OutputOption = Annotated[
    Optional[str],
    typer.Option(
        "--output",
        "-O",
        help="Path to file to write command output (stdout if not specified)",
    ),
]



from enum import Enum
from typing import List, Optional

from fastapi import Query, Request
from pydantic import BaseModel, Field


class PaginationParams(BaseModel):
    request: Request
    limit: int = Query(default=20, ge=0, le=500)
    offset: int = Query(default=0, ge=0)

    class Config:
        arbitrary_types_allowed = True


class OutputFormat(str, Enum):
    json = "json"
    tsv = "tsv"


class SemsimMetric(str, Enum):
    ANCESTOR_INFORMATION_CONTENT = "ancestor_information_content"
    # COSINE_SIMILARITY = "cosine_similarity"  # Not implemented
    JACCARD_SIMILARITY = "jaccard_similarity"
    PHENODIGM_SCORE = "phenodigm_score"

    def __str__(self):
        return self.value


class SemsimSearchGroup(Enum):
    HGNC = "Human Genes"
    MGI = "Mouse Genes"
    RGD = "Rat Genes"
    ZFIN = "Zebrafish Genes"
    WB = "C. Elegans Genes"
    MONDO = "Human Diseases"


class SemsimDirectionality(str, Enum):
    BIDIRECTIONAL = "bidirectional"
    SUBJECT_TO_OBJECT = "subject_to_object"
    OBJECT_TO_SUBJECT = "object_to_subject"


class SemsimCompareRequest(BaseModel):
    subjects: List[str] = Field(..., title="List of subjects for comparison")
    objects: List[str] = Field(..., title="List of objects for comparison")
    metric: SemsimMetric = Field(SemsimMetric.ANCESTOR_INFORMATION_CONTENT, title="Similarity metric to use")


class SemsimMultiCompareObject(BaseModel):
    id: Optional[str] = Field(None, title="ID of the object set")
    label: str = Field(..., title="Label of the object set")
    phenotypes: List[str] = Field(..., title="List of object for comparison")


class SemsimMultiCompareRequest(BaseModel):
    subjects: List[str] = Field(..., title="List of subjects for comparison")
    object_sets: List[SemsimMultiCompareObject] = Field(..., title="List of object sets for comparison")
    metric: SemsimMetric = Field(SemsimMetric.ANCESTOR_INFORMATION_CONTENT, title="Similarity metric to use")


class SemsimSearchRequest(BaseModel):
    termset: List[str] = Field(..., title="Termset to search")
    group: SemsimSearchGroup = Field(..., title="Group of entities to search within (e.g. Human Genes)")
    metric: SemsimMetric = Field(SemsimMetric.ANCESTOR_INFORMATION_CONTENT, title="Similarity metric to use")
    directionality: SemsimDirectionality = (
        Query(SemsimDirectionality.BIDIRECTIONAL, title="Directionality of the search"),
    )
    limit: Optional[int] = Field(10, title="Limit the number of results", ge=1, le=50)


class TextAnnotationRequest(BaseModel):
    content: str = Field(..., title="The text content to annotate")



from __future__ import annotations
from datetime import datetime, date, time
from decimal import Decimal
from enum import Enum
import re
import sys
from typing import Any, ClassVar, List, Literal, Dict, Optional, Union
from pydantic import BaseModel, ConfigDict, Field, RootModel, field_validator

metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        extra="allow",
        arbitrary_types_allowed=True,
        use_enum_values=True,
        strict=False,
    )
    pass


class LinkMLMeta(RootModel):
    root: Dict[str, Any] = {}
    model_config = ConfigDict(frozen=True)

    def __getattr__(self, key: str):
        return getattr(self.root, key)

    def __getitem__(self, key: str):
        return self.root[key]

    def __setitem__(self, key: str, value):
        self.root[key] = value

    def __contains__(self, key: str) -> bool:
        return key in self.root


linkml_meta = LinkMLMeta(
    {
        "default_prefix": "https://w3id.org/monarch/monarch-py/",
        "default_range": "string",
        "description": "Data models for the Monarch Initiative data access library",
        "id": "https://w3id.org/monarch/monarch-py",
        "imports": ["linkml:types", "similarity"],
        "name": "monarch-py",
        "prefixes": {
            "biolink": {"prefix_prefix": "biolink", "prefix_reference": "https://w3id.org/biolink/vocab/"},
            "linkml": {"prefix_prefix": "linkml", "prefix_reference": "https://w3id.org/linkml/"},
        },
        "source_file": "/Users/kschaper/Monarch/monarch-app/backend/src/monarch_py/datamodels/model.yaml",
    }
)


class AssociationDirectionEnum(str, Enum):
    """
    The directionality of an association as it relates to a specified entity, with edges being categorized as incoming or outgoing
    """

    # An association for which a specified entity is the object or part of the object closure
    incoming = "incoming"
    # An association for which a specified entity is the subject or part of the subject closure
    outgoing = "outgoing"


class PairwiseSimilarity(ConfiguredBaseModel):
    """
    Abstract grouping for representing individual pairwise similarities
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {"abstract": True, "from_schema": "https://w3id.org/monarch/monarch-py-similarity"}
    )

    pass


class TermPairwiseSimilarity(PairwiseSimilarity):
    """
    A simple pairwise similarity between two atomic concepts/terms
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py-similarity"})

    subject_id: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "subject_id", "domain_of": ["TermPairwiseSimilarity", "Mapping"]}},
    )
    subject_label: Optional[str] = Field(
        None,
        description="""The name of the subject entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    subject_source: Optional[str] = Field(
        None,
        description="""the source for the first entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_source",
                "domain_of": ["TermPairwiseSimilarity"],
                "slot_uri": "sssom:subject_source",
            }
        },
    )
    object_id: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "object_id", "domain_of": ["TermPairwiseSimilarity", "Mapping"]}},
    )
    object_label: Optional[str] = Field(
        None,
        description="""The name of the object entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    object_source: Optional[str] = Field(
        None,
        description="""the source for the second entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_source",
                "domain_of": ["TermPairwiseSimilarity"],
                "slot_uri": "sssom:object_source",
            }
        },
    )
    ancestor_id: Optional[str] = Field(
        None,
        description="""the most recent common ancestor of the two compared entities. If there are multiple MRCAs then the most informative one is selected""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "ancestor_id",
                "domain_of": ["TermPairwiseSimilarity"],
                "todos": ["decide on what to do when there are multiple possible ancestos"],
            }
        },
    )
    ancestor_label: Optional[str] = Field(
        None,
        description="""the name or label of the ancestor concept""",
        json_schema_extra={"linkml_meta": {"alias": "ancestor_label", "domain_of": ["TermPairwiseSimilarity"]}},
    )
    ancestor_source: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "ancestor_source", "domain_of": ["TermPairwiseSimilarity"]}}
    )
    object_information_content: Optional[float] = Field(
        None,
        description="""The IC of the object""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_information_content",
                "domain_of": ["TermPairwiseSimilarity"],
                "is_a": "information_content",
            }
        },
    )
    subject_information_content: Optional[float] = Field(
        None,
        description="""The IC of the subject""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_information_content",
                "domain_of": ["TermPairwiseSimilarity"],
                "is_a": "information_content",
            }
        },
    )
    ancestor_information_content: Optional[float] = Field(
        None,
        description="""The IC of the object""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "ancestor_information_content",
                "domain_of": ["TermPairwiseSimilarity"],
                "is_a": "information_content",
            }
        },
    )
    jaccard_similarity: Optional[float] = Field(
        None,
        description="""The number of concepts in the intersection divided by the number in the union""",
        json_schema_extra={
            "linkml_meta": {"alias": "jaccard_similarity", "domain_of": ["TermPairwiseSimilarity"], "is_a": "score"}
        },
    )
    cosine_similarity: Optional[float] = Field(
        None,
        description="""the dot product of two node embeddings divided by the product of their lengths""",
        json_schema_extra={
            "linkml_meta": {"alias": "cosine_similarity", "domain_of": ["TermPairwiseSimilarity"], "is_a": "score"}
        },
    )
    dice_similarity: Optional[float] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "dice_similarity", "domain_of": ["TermPairwiseSimilarity"], "is_a": "score"}
        },
    )
    phenodigm_score: Optional[float] = Field(
        None,
        description="""the geometric mean of the jaccard similarity and the information content""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "phenodigm_score",
                "domain_of": ["TermPairwiseSimilarity"],
                "equals_expression": "sqrt({jaccard_similarity} * {information_content})",
                "is_a": "score",
            }
        },
    )


class TermSetPairwiseSimilarity(PairwiseSimilarity):
    """
    A simple pairwise similarity between two sets of concepts/terms
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py-similarity"})

    subject_termset: Optional[Dict[str, Union[str, TermInfo]]] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "subject_termset", "domain_of": ["TermSetPairwiseSimilarity"]}},
    )
    object_termset: Optional[Dict[str, Union[str, TermInfo]]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "object_termset", "domain_of": ["TermSetPairwiseSimilarity"]}}
    )
    subject_best_matches: Optional[Dict[str, BestMatch]] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "subject_best_matches", "domain_of": ["TermSetPairwiseSimilarity"]}
        },
    )
    object_best_matches: Optional[Dict[str, BestMatch]] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "object_best_matches", "domain_of": ["TermSetPairwiseSimilarity"]}},
    )
    average_score: Optional[float] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "average_score", "domain_of": ["TermSetPairwiseSimilarity"]}}
    )
    best_score: Optional[float] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "best_score", "domain_of": ["TermSetPairwiseSimilarity"]}}
    )
    metric: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "metric", "domain_of": ["TermSetPairwiseSimilarity"]}}
    )


class TermInfo(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py-similarity"})

    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    label: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {
                "alias": "label",
                "domain_of": ["TermInfo", "FacetValue", "FacetField"],
                "slot_uri": "rdfs:label",
            }
        },
    )


class BestMatch(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py-similarity"})

    match_source: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "match_source",
                "comments": ["note that the match_source is either the subject or the object"],
                "domain_of": ["BestMatch"],
            }
        },
    )
    match_source_label: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "match_source_label", "domain_of": ["BestMatch"]}}
    )
    match_target: Optional[str] = Field(
        None,
        description="""the entity matches""",
        json_schema_extra={"linkml_meta": {"alias": "match_target", "domain_of": ["BestMatch"]}},
    )
    match_target_label: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "match_target_label", "domain_of": ["BestMatch"]}}
    )
    score: float = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {"alias": "score", "domain_of": ["BestMatch", "SemsimSearchResult", "SearchResult"]}
        },
    )
    match_subsumer: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "match_subsumer", "domain_of": ["BestMatch"]}}
    )
    match_subsumer_label: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "match_subsumer_label", "domain_of": ["BestMatch"]}}
    )
    similarity: TermPairwiseSimilarity = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "similarity", "domain_of": ["BestMatch", "SemsimSearchResult"]}},
    )


class SemsimSearchResult(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py-similarity",
            "slot_usage": {"subject": {"inlined": True, "name": "subject", "range": "Entity"}},
        }
    )

    subject: Entity = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject",
                "domain_of": ["SemsimSearchResult", "Association", "CompactAssociation"],
            }
        },
    )
    score: Optional[float] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "score", "domain_of": ["BestMatch", "SemsimSearchResult", "SearchResult"]}
        },
    )
    similarity: Optional[TermSetPairwiseSimilarity] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "similarity", "domain_of": ["BestMatch", "SemsimSearchResult"]}},
    )


class Association(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    category: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )
    subject: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject",
                "domain_of": ["SemsimSearchResult", "Association", "CompactAssociation"],
            }
        },
    )
    original_subject: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "original_subject", "domain_of": ["Association"]}}
    )
    subject_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the subject entity""",
        json_schema_extra={"linkml_meta": {"alias": "subject_namespace", "domain_of": ["Association"]}},
    )
    subject_category: Optional[str] = Field(
        None,
        description="""The category of the subject entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    subject_closure: Optional[List[str]] = Field(
        None,
        description="""Field containing subject id and the ids of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "subject_closure", "domain_of": ["Association"]}},
    )
    subject_label: Optional[str] = Field(
        None,
        description="""The name of the subject entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    subject_closure_label: Optional[List[str]] = Field(
        None,
        description="""Field containing subject name and the names of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "subject_closure_label", "domain_of": ["Association"]}},
    )
    subject_taxon: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "subject_taxon", "domain_of": ["Association"], "is_a": "in_taxon"}},
    )
    subject_taxon_label: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "subject_taxon_label", "domain_of": ["Association"], "is_a": "in_taxon_label"}
        },
    )
    predicate: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "predicate", "domain_of": ["Association", "CompactAssociation"]}},
    )
    original_predicate: Optional[str] = Field(
        None,
        description="""used to hold the original relation/predicate that an external knowledge source uses before transformation to match the biolink-model specification.""",
        json_schema_extra={"linkml_meta": {"alias": "original_predicate", "domain_of": ["Association"]}},
    )
    object: str = Field(
        ..., json_schema_extra={"linkml_meta": {"alias": "object", "domain_of": ["Association", "CompactAssociation"]}}
    )
    original_object: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "original_object", "domain_of": ["Association"]}}
    )
    object_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the object entity""",
        json_schema_extra={"linkml_meta": {"alias": "object_namespace", "domain_of": ["Association"]}},
    )
    object_category: Optional[str] = Field(
        None,
        description="""The category of the object entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    object_closure: Optional[List[str]] = Field(
        None,
        description="""Field containing object id and the ids of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "object_closure", "domain_of": ["Association"]}},
    )
    object_label: Optional[str] = Field(
        None,
        description="""The name of the object entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    object_closure_label: Optional[List[str]] = Field(
        None,
        description="""Field containing object name and the names of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "object_closure_label", "domain_of": ["Association"]}},
    )
    object_taxon: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "object_taxon", "domain_of": ["Association"], "is_a": "in_taxon"}},
    )
    object_taxon_label: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "object_taxon_label", "domain_of": ["Association"], "is_a": "in_taxon_label"}
        },
    )
    primary_knowledge_source: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "primary_knowledge_source", "domain_of": ["Association"]}}
    )
    aggregator_knowledge_source: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "aggregator_knowledge_source", "domain_of": ["Association"]}}
    )
    negated: Optional[bool] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "negated", "domain_of": ["Association", "CompactAssociation"]}},
    )
    pathway: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "pathway", "domain_of": ["Association"]}}
    )
    evidence_count: Optional[int] = Field(
        None,
        description="""count of supporting documents, evidence codes, and sources supplying evidence""",
        json_schema_extra={"linkml_meta": {"alias": "evidence_count", "domain_of": ["Association"]}},
    )
    knowledge_level: str = Field(
        ...,
        description="""Describes the level of knowledge expressed in a statement, based on the reasoning or analysis methods used to generate the statement, or the scope or specificity of what the statement expresses to be true.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "knowledge_level",
                "domain_of": ["Association"],
                "notes": [
                    "The range in this schema is represented as a string, but is "
                    "constrained  to values from biolink:KnowledgeLevelEnum at ingest "
                    "time"
                ],
                "slot_uri": "biolink:knowledge_level",
            }
        },
    )
    agent_type: str = Field(
        ...,
        description="""Describes the high-level category of agent who originally generated a  statement of knowledge or other type of information.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "agent_type",
                "domain_of": ["Association"],
                "notes": [
                    "The range in this schema is represented as a string, but is "
                    "constrained  to values from biolink:AgentTypeEnum at ingest time"
                ],
                "slot_uri": "biolink:agent_type",
            }
        },
    )
    has_evidence: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "has_evidence", "domain_of": ["Association"]}}
    )
    has_evidence_links: Optional[List[ExpandedCurie]] = Field(
        None,
        description="""List of ExpandedCuries with id and url for evidence""",
        json_schema_extra={"linkml_meta": {"alias": "has_evidence_links", "domain_of": ["Association"]}},
    )
    has_count: Optional[int] = Field(
        None,
        description="""count of out of has_total representing a frequency""",
        json_schema_extra={"linkml_meta": {"alias": "has_count", "domain_of": ["Association"]}},
    )
    has_total: Optional[int] = Field(
        None,
        description="""total, devided by has_count, representing a frequency""",
        json_schema_extra={"linkml_meta": {"alias": "has_total", "domain_of": ["Association"]}},
    )
    has_percentage: Optional[float] = Field(
        None,
        description="""percentage, which may be calculated from has_count and has_total, as 100 * quotient or provided directly, rounded to the integer level""",
        json_schema_extra={"linkml_meta": {"alias": "has_percentage", "domain_of": ["Association"]}},
    )
    has_quotient: Optional[float] = Field(
        None,
        description="""quotient, which should be 1/100 of has_percentage""",
        json_schema_extra={"linkml_meta": {"alias": "has_quotient", "domain_of": ["Association"]}},
    )
    grouping_key: Optional[str] = Field(
        None,
        description="""A concatenation of fields used to group associations with the same essential/defining properties""",
        json_schema_extra={"linkml_meta": {"alias": "grouping_key", "domain_of": ["Association"]}},
    )
    provided_by: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "provided_by", "domain_of": ["Association", "Entity"]}}
    )
    provided_by_link: Optional[ExpandedCurie] = Field(
        None,
        description="""A link to the docs for the knowledge source that provided the node/edge.""",
        json_schema_extra={"linkml_meta": {"alias": "provided_by_link", "domain_of": ["Association", "Node"]}},
    )
    publications: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "publications", "domain_of": ["Association"]}}
    )
    publications_links: Optional[List[ExpandedCurie]] = Field(
        None,
        description="""List of ExpandedCuries with id and url for publications""",
        json_schema_extra={"linkml_meta": {"alias": "publications_links", "domain_of": ["Association"]}},
    )
    frequency_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "frequency_qualifier", "domain_of": ["Association"]}}
    )
    onset_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "onset_qualifier", "domain_of": ["Association"]}}
    )
    sex_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "sex_qualifier", "domain_of": ["Association"]}}
    )
    stage_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "stage_qualifier", "domain_of": ["Association"]}}
    )
    qualifiers: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "qualifiers", "domain_of": ["Association"]}}
    )
    qualifiers_label: Optional[str] = Field(
        None,
        description="""The name of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifiers_label", "domain_of": ["Association"], "is_a": "name"}},
    )
    qualifiers_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifiers_namespace", "domain_of": ["Association"]}},
    )
    qualifiers_category: Optional[str] = Field(
        None,
        description="""The category of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "qualifiers_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    qualifier: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "qualifier", "domain_of": ["Association"]}}
    )
    qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifier_label", "domain_of": ["Association"], "is_a": "name"}},
    )
    qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifier_namespace", "domain_of": ["Association"]}},
    )
    qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    frequency_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "frequency_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    frequency_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "frequency_qualifier_namespace", "domain_of": ["Association"]}},
    )
    frequency_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "frequency_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    onset_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the onset_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "onset_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    onset_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the onset_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "onset_qualifier_namespace", "domain_of": ["Association"]}},
    )
    onset_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the onset_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "onset_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    sex_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the sex_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "sex_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    sex_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the sex_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "sex_qualifier_namespace", "domain_of": ["Association"]}},
    )
    sex_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the sex_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "sex_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    stage_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the stage_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "stage_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    stage_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the stage_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "stage_qualifier_namespace", "domain_of": ["Association"]}},
    )
    stage_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the stage_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "stage_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    disease_context_qualifier: Optional[str] = Field(
        None,
        description="""A context qualifier representing a disease or condition in which a relationship expressed in an association took place.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "disease_context_qualifier",
                "domain_of": ["Association"],
                "examples": [{"value": "MONDO:0004979"}, {"value": "MONDO:0005148"}],
            }
        },
    )
    disease_context_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the disease_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "disease_context_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    disease_context_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the disease_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "disease_context_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    disease_context_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the disease_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "disease_context_qualifier_category",
                "domain_of": ["Association"],
                "is_a": "category",
            }
        },
    )
    disease_context_qualifier_closure: Optional[List[str]] = Field(
        None,
        description="""Field containing disease_context_qualifier id and the ids of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "disease_context_qualifier_closure", "domain_of": ["Association"]}},
    )
    disease_context_qualifier_closure_label: Optional[List[str]] = Field(
        None,
        description="""Field containing disease_context_qualifier name and the names of all of it's ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "disease_context_qualifier_closure_label", "domain_of": ["Association"]}
        },
    )
    species_context_qualifier: Optional[str] = Field(
        None,
        description="""A context qualifier representing a species in which a relationship expressed in an association took place.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "species_context_qualifier",
                "domain_of": ["Association"],
                "examples": [{"value": "NCBITaxon:9606"}],
            }
        },
    )
    species_context_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the species_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "species_context_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    species_context_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the species_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "species_context_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    species_context_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the species_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "species_context_qualifier_category",
                "domain_of": ["Association"],
                "is_a": "category",
            }
        },
    )
    subject_specialization_qualifier: Optional[str] = Field(
        None,
        description="""A qualifier that composes with a core subject/object concept to define a more specific version of the subject concept, specifically using an ontology term that is not a subclass or descendant of the core concept and in the vast majority of cases, is of a different ontological namespace than the category or namespace of the subject identifier.""",
        json_schema_extra={"linkml_meta": {"alias": "subject_specialization_qualifier", "domain_of": ["Association"]}},
    )
    subject_specialization_qualifier_label: Optional[str] = Field(
        None,
        description="""A label for the subject_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_label", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the subject_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the subject_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_category", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_closure: Optional[str] = Field(
        None,
        description="""A closure of the subject_specialization_qualifier, including the subject_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_closure", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_closure_label: Optional[str] = Field(
        None,
        description="""A closure of the subject_specialization_qualifier, including the subject_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_closure_label", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier: Optional[str] = Field(
        None,
        description="""A qualifier that composes with a core subject/object concept to define a more specific version of the object concept, specifically using an ontology term that is not a subclass or descendant of the core concept and in the vast majority of cases, is of a different ontological namespace than the category or namespace of the object identifier.""",
        json_schema_extra={"linkml_meta": {"alias": "object_specialization_qualifier", "domain_of": ["Association"]}},
    )
    object_specialization_qualifier_label: Optional[str] = Field(
        None,
        description="""A label for the object_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_label", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the object_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the object_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_category", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_closure: Optional[str] = Field(
        None,
        description="""A closure of the object_specialization_qualifier, including the object_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_closure", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_closure_label: Optional[str] = Field(
        None,
        description="""A closure of the object_specialization_qualifier, including the object_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_closure_label", "domain_of": ["Association"]}
        },
    )


class AssociationCountList(ConfiguredBaseModel):
    """
    Container class for a list of association counts
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "AssociationCount"}},
        }
    )

    items: List[AssociationCount] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )


class CompactAssociation(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    category: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )
    subject: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject",
                "domain_of": ["SemsimSearchResult", "Association", "CompactAssociation"],
            }
        },
    )
    subject_label: Optional[str] = Field(
        None,
        description="""The name of the subject entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    predicate: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "predicate", "domain_of": ["Association", "CompactAssociation"]}},
    )
    object: str = Field(
        ..., json_schema_extra={"linkml_meta": {"alias": "object", "domain_of": ["Association", "CompactAssociation"]}}
    )
    object_label: Optional[str] = Field(
        None,
        description="""The name of the object entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    negated: Optional[bool] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "negated", "domain_of": ["Association", "CompactAssociation"]}},
    )


class AssociationTypeMapping(ConfiguredBaseModel):
    """
    A data class to hold the necessary information to produce association type counts for given  entities with appropriate directional labels
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {
                "category": {
                    "description": "The biolink category to use in " "queries for this association type",
                    "multivalued": False,
                    "name": "category",
                    "required": True,
                },
                "object_label": {
                    "description": "A label to describe the "
                    "objects of the association "
                    "type as a whole for use in "
                    "the UI",
                    "name": "object_label",
                },
                "subject_label": {
                    "description": "A label to describe the "
                    "subjects of the association "
                    "type as a whole for use in "
                    "the UI",
                    "name": "subject_label",
                },
                "symmetric": {
                    "description": "Whether the association type is "
                    "symmetric, meaning that the "
                    "subject and object labels should "
                    "be interchangeable",
                    "ifabsent": "False",
                    "name": "symmetric",
                    "required": True,
                },
            },
        }
    )

    subject_label: Optional[str] = Field(
        None,
        description="""A label to describe the subjects of the association type as a whole for use in the UI""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    object_label: Optional[str] = Field(
        None,
        description="""A label to describe the objects of the association type as a whole for use in the UI""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    symmetric: bool = Field(
        False,
        description="""Whether the association type is symmetric, meaning that the subject and object labels should be interchangeable""",
        json_schema_extra={
            "linkml_meta": {"alias": "symmetric", "domain_of": ["AssociationTypeMapping"], "ifabsent": "False"}
        },
    )
    category: str = Field(
        ...,
        description="""The biolink category to use in queries for this association type""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )


class DirectionalAssociation(Association):
    """
    An association that gives it's direction relative to a specified entity
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    direction: AssociationDirectionEnum = Field(
        ...,
        description="""The directionality of the association relative to a given entity for an association_count. If the entity is the subject or in the subject closure, the direction is forwards, if it is the object or in the object closure, the direction is backwards.""",
        json_schema_extra={"linkml_meta": {"alias": "direction", "domain_of": ["DirectionalAssociation"]}},
    )
    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    category: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )
    subject: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject",
                "domain_of": ["SemsimSearchResult", "Association", "CompactAssociation"],
            }
        },
    )
    original_subject: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "original_subject", "domain_of": ["Association"]}}
    )
    subject_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the subject entity""",
        json_schema_extra={"linkml_meta": {"alias": "subject_namespace", "domain_of": ["Association"]}},
    )
    subject_category: Optional[str] = Field(
        None,
        description="""The category of the subject entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    subject_closure: Optional[List[str]] = Field(
        None,
        description="""Field containing subject id and the ids of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "subject_closure", "domain_of": ["Association"]}},
    )
    subject_label: Optional[str] = Field(
        None,
        description="""The name of the subject entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    subject_closure_label: Optional[List[str]] = Field(
        None,
        description="""Field containing subject name and the names of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "subject_closure_label", "domain_of": ["Association"]}},
    )
    subject_taxon: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "subject_taxon", "domain_of": ["Association"], "is_a": "in_taxon"}},
    )
    subject_taxon_label: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "subject_taxon_label", "domain_of": ["Association"], "is_a": "in_taxon_label"}
        },
    )
    predicate: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "predicate", "domain_of": ["Association", "CompactAssociation"]}},
    )
    original_predicate: Optional[str] = Field(
        None,
        description="""used to hold the original relation/predicate that an external knowledge source uses before transformation to match the biolink-model specification.""",
        json_schema_extra={"linkml_meta": {"alias": "original_predicate", "domain_of": ["Association"]}},
    )
    object: str = Field(
        ..., json_schema_extra={"linkml_meta": {"alias": "object", "domain_of": ["Association", "CompactAssociation"]}}
    )
    original_object: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "original_object", "domain_of": ["Association"]}}
    )
    object_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the object entity""",
        json_schema_extra={"linkml_meta": {"alias": "object_namespace", "domain_of": ["Association"]}},
    )
    object_category: Optional[str] = Field(
        None,
        description="""The category of the object entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    object_closure: Optional[List[str]] = Field(
        None,
        description="""Field containing object id and the ids of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "object_closure", "domain_of": ["Association"]}},
    )
    object_label: Optional[str] = Field(
        None,
        description="""The name of the object entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    object_closure_label: Optional[List[str]] = Field(
        None,
        description="""Field containing object name and the names of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "object_closure_label", "domain_of": ["Association"]}},
    )
    object_taxon: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "object_taxon", "domain_of": ["Association"], "is_a": "in_taxon"}},
    )
    object_taxon_label: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "object_taxon_label", "domain_of": ["Association"], "is_a": "in_taxon_label"}
        },
    )
    primary_knowledge_source: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "primary_knowledge_source", "domain_of": ["Association"]}}
    )
    aggregator_knowledge_source: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "aggregator_knowledge_source", "domain_of": ["Association"]}}
    )
    negated: Optional[bool] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "negated", "domain_of": ["Association", "CompactAssociation"]}},
    )
    pathway: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "pathway", "domain_of": ["Association"]}}
    )
    evidence_count: Optional[int] = Field(
        None,
        description="""count of supporting documents, evidence codes, and sources supplying evidence""",
        json_schema_extra={"linkml_meta": {"alias": "evidence_count", "domain_of": ["Association"]}},
    )
    knowledge_level: str = Field(
        ...,
        description="""Describes the level of knowledge expressed in a statement, based on the reasoning or analysis methods used to generate the statement, or the scope or specificity of what the statement expresses to be true.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "knowledge_level",
                "domain_of": ["Association"],
                "notes": [
                    "The range in this schema is represented as a string, but is "
                    "constrained  to values from biolink:KnowledgeLevelEnum at ingest "
                    "time"
                ],
                "slot_uri": "biolink:knowledge_level",
            }
        },
    )
    agent_type: str = Field(
        ...,
        description="""Describes the high-level category of agent who originally generated a  statement of knowledge or other type of information.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "agent_type",
                "domain_of": ["Association"],
                "notes": [
                    "The range in this schema is represented as a string, but is "
                    "constrained  to values from biolink:AgentTypeEnum at ingest time"
                ],
                "slot_uri": "biolink:agent_type",
            }
        },
    )
    has_evidence: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "has_evidence", "domain_of": ["Association"]}}
    )
    has_evidence_links: Optional[List[ExpandedCurie]] = Field(
        None,
        description="""List of ExpandedCuries with id and url for evidence""",
        json_schema_extra={"linkml_meta": {"alias": "has_evidence_links", "domain_of": ["Association"]}},
    )
    has_count: Optional[int] = Field(
        None,
        description="""count of out of has_total representing a frequency""",
        json_schema_extra={"linkml_meta": {"alias": "has_count", "domain_of": ["Association"]}},
    )
    has_total: Optional[int] = Field(
        None,
        description="""total, devided by has_count, representing a frequency""",
        json_schema_extra={"linkml_meta": {"alias": "has_total", "domain_of": ["Association"]}},
    )
    has_percentage: Optional[float] = Field(
        None,
        description="""percentage, which may be calculated from has_count and has_total, as 100 * quotient or provided directly, rounded to the integer level""",
        json_schema_extra={"linkml_meta": {"alias": "has_percentage", "domain_of": ["Association"]}},
    )
    has_quotient: Optional[float] = Field(
        None,
        description="""quotient, which should be 1/100 of has_percentage""",
        json_schema_extra={"linkml_meta": {"alias": "has_quotient", "domain_of": ["Association"]}},
    )
    grouping_key: Optional[str] = Field(
        None,
        description="""A concatenation of fields used to group associations with the same essential/defining properties""",
        json_schema_extra={"linkml_meta": {"alias": "grouping_key", "domain_of": ["Association"]}},
    )
    provided_by: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "provided_by", "domain_of": ["Association", "Entity"]}}
    )
    provided_by_link: Optional[ExpandedCurie] = Field(
        None,
        description="""A link to the docs for the knowledge source that provided the node/edge.""",
        json_schema_extra={"linkml_meta": {"alias": "provided_by_link", "domain_of": ["Association", "Node"]}},
    )
    publications: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "publications", "domain_of": ["Association"]}}
    )
    publications_links: Optional[List[ExpandedCurie]] = Field(
        None,
        description="""List of ExpandedCuries with id and url for publications""",
        json_schema_extra={"linkml_meta": {"alias": "publications_links", "domain_of": ["Association"]}},
    )
    frequency_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "frequency_qualifier", "domain_of": ["Association"]}}
    )
    onset_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "onset_qualifier", "domain_of": ["Association"]}}
    )
    sex_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "sex_qualifier", "domain_of": ["Association"]}}
    )
    stage_qualifier: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "stage_qualifier", "domain_of": ["Association"]}}
    )
    qualifiers: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "qualifiers", "domain_of": ["Association"]}}
    )
    qualifiers_label: Optional[str] = Field(
        None,
        description="""The name of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifiers_label", "domain_of": ["Association"], "is_a": "name"}},
    )
    qualifiers_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifiers_namespace", "domain_of": ["Association"]}},
    )
    qualifiers_category: Optional[str] = Field(
        None,
        description="""The category of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "qualifiers_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    qualifier: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "qualifier", "domain_of": ["Association"]}}
    )
    qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifier_label", "domain_of": ["Association"], "is_a": "name"}},
    )
    qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "qualifier_namespace", "domain_of": ["Association"]}},
    )
    qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    frequency_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "frequency_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    frequency_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the frequency_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "frequency_qualifier_namespace", "domain_of": ["Association"]}},
    )
    frequency_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the frequency_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "frequency_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    onset_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the onset_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "onset_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    onset_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the onset_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "onset_qualifier_namespace", "domain_of": ["Association"]}},
    )
    onset_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the onset_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "onset_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    sex_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the sex_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "sex_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    sex_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the sex_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "sex_qualifier_namespace", "domain_of": ["Association"]}},
    )
    sex_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the sex_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "sex_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    stage_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the stage_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "stage_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    stage_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the stage_qualifier entity""",
        json_schema_extra={"linkml_meta": {"alias": "stage_qualifier_namespace", "domain_of": ["Association"]}},
    )
    stage_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the stage_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "stage_qualifier_category", "domain_of": ["Association"], "is_a": "category"}
        },
    )
    disease_context_qualifier: Optional[str] = Field(
        None,
        description="""A context qualifier representing a disease or condition in which a relationship expressed in an association took place.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "disease_context_qualifier",
                "domain_of": ["Association"],
                "examples": [{"value": "MONDO:0004979"}, {"value": "MONDO:0005148"}],
            }
        },
    )
    disease_context_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the disease_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "disease_context_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    disease_context_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the disease_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "disease_context_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    disease_context_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the disease_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "disease_context_qualifier_category",
                "domain_of": ["Association"],
                "is_a": "category",
            }
        },
    )
    disease_context_qualifier_closure: Optional[List[str]] = Field(
        None,
        description="""Field containing disease_context_qualifier id and the ids of all of it's ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "disease_context_qualifier_closure", "domain_of": ["Association"]}},
    )
    disease_context_qualifier_closure_label: Optional[List[str]] = Field(
        None,
        description="""Field containing disease_context_qualifier name and the names of all of it's ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "disease_context_qualifier_closure_label", "domain_of": ["Association"]}
        },
    )
    species_context_qualifier: Optional[str] = Field(
        None,
        description="""A context qualifier representing a species in which a relationship expressed in an association took place.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "species_context_qualifier",
                "domain_of": ["Association"],
                "examples": [{"value": "NCBITaxon:9606"}],
            }
        },
    )
    species_context_qualifier_label: Optional[str] = Field(
        None,
        description="""The name of the species_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "species_context_qualifier_label", "domain_of": ["Association"], "is_a": "name"}
        },
    )
    species_context_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the species_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {"alias": "species_context_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    species_context_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the species_context_qualifier entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "species_context_qualifier_category",
                "domain_of": ["Association"],
                "is_a": "category",
            }
        },
    )
    subject_specialization_qualifier: Optional[str] = Field(
        None,
        description="""A qualifier that composes with a core subject/object concept to define a more specific version of the subject concept, specifically using an ontology term that is not a subclass or descendant of the core concept and in the vast majority of cases, is of a different ontological namespace than the category or namespace of the subject identifier.""",
        json_schema_extra={"linkml_meta": {"alias": "subject_specialization_qualifier", "domain_of": ["Association"]}},
    )
    subject_specialization_qualifier_label: Optional[str] = Field(
        None,
        description="""A label for the subject_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_label", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the subject_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the subject_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_category", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_closure: Optional[str] = Field(
        None,
        description="""A closure of the subject_specialization_qualifier, including the subject_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_closure", "domain_of": ["Association"]}
        },
    )
    subject_specialization_qualifier_closure_label: Optional[str] = Field(
        None,
        description="""A closure of the subject_specialization_qualifier, including the subject_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "subject_specialization_qualifier_closure_label", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier: Optional[str] = Field(
        None,
        description="""A qualifier that composes with a core subject/object concept to define a more specific version of the object concept, specifically using an ontology term that is not a subclass or descendant of the core concept and in the vast majority of cases, is of a different ontological namespace than the category or namespace of the object identifier.""",
        json_schema_extra={"linkml_meta": {"alias": "object_specialization_qualifier", "domain_of": ["Association"]}},
    )
    object_specialization_qualifier_label: Optional[str] = Field(
        None,
        description="""A label for the object_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_label", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix of the object_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_namespace", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_category: Optional[str] = Field(
        None,
        description="""The category of the object_specialization_qualifier""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_category", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_closure: Optional[str] = Field(
        None,
        description="""A closure of the object_specialization_qualifier, including the object_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_closure", "domain_of": ["Association"]}
        },
    )
    object_specialization_qualifier_closure_label: Optional[str] = Field(
        None,
        description="""A closure of the object_specialization_qualifier, including the object_specialization_qualifier itself and all of its ancestors""",
        json_schema_extra={
            "linkml_meta": {"alias": "object_specialization_qualifier_closure_label", "domain_of": ["Association"]}
        },
    )


class ExpandedCurie(ConfiguredBaseModel):
    """
    A curie bundled along with its expanded url
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    url: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "url", "domain_of": ["ExpandedCurie", "Release"]}}
    )


class Entity(ConfiguredBaseModel):
    """
    Represents an Entity in the Monarch KG data model
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    category: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )
    name: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "name", "domain_of": ["Entity", "MultiEntityAssociationResults"]}},
    )
    full_name: Optional[str] = Field(
        None,
        description="""The long form name of an entity""",
        json_schema_extra={"linkml_meta": {"alias": "full_name", "domain_of": ["Entity"]}},
    )
    deprecated: Optional[bool] = Field(
        None,
        description="""A boolean flag indicating that an entity is no longer considered current or valid.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "deprecated",
                "domain_of": ["Entity"],
                "exact_mappings": ["oboInOwl:ObsoleteClass"],
            }
        },
    )
    description: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "description", "domain_of": ["Entity"]}}
    )
    xref: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "xref", "domain_of": ["Entity"]}}
    )
    provided_by: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "provided_by", "domain_of": ["Association", "Entity"]}}
    )
    in_taxon: Optional[str] = Field(
        None,
        description="""The biolink taxon that the entity is in the closure of.""",
        json_schema_extra={"linkml_meta": {"alias": "in_taxon", "domain_of": ["Entity", "Node"]}},
    )
    in_taxon_label: Optional[str] = Field(
        None,
        description="""The label of the biolink taxon that the entity is in the closure of.""",
        json_schema_extra={"linkml_meta": {"alias": "in_taxon_label", "domain_of": ["Entity", "Node"]}},
    )
    symbol: Optional[str] = Field(None, json_schema_extra={"linkml_meta": {"alias": "symbol", "domain_of": ["Entity"]}})
    synonym: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "synonym", "domain_of": ["Entity"]}}
    )
    uri: Optional[str] = Field(
        None,
        description="""The URI of the entity""",
        json_schema_extra={"linkml_meta": {"alias": "uri", "domain_of": ["Entity"]}},
    )
    iri: Optional[str] = Field(None, json_schema_extra={"linkml_meta": {"alias": "iri", "domain_of": ["Entity"]}})
    namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix portion of this entity's identifier""",
        json_schema_extra={"linkml_meta": {"alias": "namespace", "domain_of": ["Entity"]}},
    )
    has_phenotype: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype identifiers that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype", "domain_of": ["Entity"]}},
    )
    has_phenotype_label: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype labels that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_label", "domain_of": ["Entity"]}},
    )
    has_phenotype_closure: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype identifiers that are known to be associated with this entity expanded to include all ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_closure", "domain_of": ["Entity"]}},
    )
    has_phenotype_closure_label: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype labels that are known to be associated with this entity expanded to include all ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_closure_label", "domain_of": ["Entity"]}},
    )
    has_phenotype_count: Optional[int] = Field(
        None,
        description="""A count of the number of phenotypes that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_count", "domain_of": ["Entity"]}},
    )


class FacetValue(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    label: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "label", "domain_of": ["TermInfo", "FacetValue", "FacetField"]}},
    )
    count: Optional[int] = Field(
        None,
        description="""count of documents""",
        json_schema_extra={"linkml_meta": {"alias": "count", "domain_of": ["FacetValue"]}},
    )


class AssociationCount(FacetValue):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"category": {"multivalued": False, "name": "category"}},
        }
    )

    category: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )
    label: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "label", "domain_of": ["TermInfo", "FacetValue", "FacetField"]}},
    )
    count: Optional[int] = Field(
        None,
        description="""count of documents""",
        json_schema_extra={"linkml_meta": {"alias": "count", "domain_of": ["FacetValue"]}},
    )


class FacetField(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    label: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "label", "domain_of": ["TermInfo", "FacetValue", "FacetField"]}},
    )
    facet_values: Optional[List[FacetValue]] = Field(
        None,
        description="""Collection of FacetValue label/value instances belonging to a FacetField""",
        json_schema_extra={"linkml_meta": {"alias": "facet_values", "domain_of": ["FacetField"]}},
    )


class HistoPheno(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "HistoBin"}},
        }
    )

    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    items: List[HistoBin] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )


class HistoBin(FacetValue):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    label: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "label", "domain_of": ["TermInfo", "FacetValue", "FacetField"]}},
    )
    count: Optional[int] = Field(
        None,
        description="""count of documents""",
        json_schema_extra={"linkml_meta": {"alias": "count", "domain_of": ["FacetValue"]}},
    )


class Mapping(ConfiguredBaseModel):
    """
    A minimal class to hold a SSSOM mapping
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    subject_id: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "subject_id", "domain_of": ["TermPairwiseSimilarity", "Mapping"]}},
    )
    subject_label: Optional[str] = Field(
        None,
        description="""The name of the subject entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "subject_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    predicate_id: str = Field(
        ..., json_schema_extra={"linkml_meta": {"alias": "predicate_id", "domain_of": ["Mapping"]}}
    )
    object_id: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "object_id", "domain_of": ["TermPairwiseSimilarity", "Mapping"]}},
    )
    object_label: Optional[str] = Field(
        None,
        description="""The name of the object entity""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "object_label",
                "domain_of": [
                    "TermPairwiseSimilarity",
                    "Association",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Mapping",
                ],
                "is_a": "name",
            }
        },
    )
    mapping_justification: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "mapping_justification", "domain_of": ["Mapping"]}}
    )
    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )


class Node(Entity):
    """
    UI container class extending Entity with additional information
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    in_taxon: Optional[str] = Field(
        None,
        description="""The biolink taxon that the entity is in the closure of.""",
        json_schema_extra={"linkml_meta": {"alias": "in_taxon", "domain_of": ["Entity", "Node"]}},
    )
    in_taxon_label: Optional[str] = Field(
        None,
        description="""The label of the biolink taxon that the entity is in the closure of.""",
        json_schema_extra={"linkml_meta": {"alias": "in_taxon_label", "domain_of": ["Entity", "Node"]}},
    )
    inheritance: Optional[Entity] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "inheritance", "domain_of": ["Node"]}}
    )
    causal_gene: Optional[List[Entity]] = Field(
        None,
        description="""A list of genes that are known to be causally associated with a disease""",
        json_schema_extra={"linkml_meta": {"alias": "causal_gene", "domain_of": ["Node"]}},
    )
    causes_disease: Optional[List[Entity]] = Field(
        None,
        description="""A list of diseases that are known to be causally associated with a gene""",
        json_schema_extra={"linkml_meta": {"alias": "causes_disease", "domain_of": ["Node"]}},
    )
    mappings: Optional[List[ExpandedCurie]] = Field(
        None,
        description="""List of ExpandedCuries with id and url for mapped entities""",
        json_schema_extra={"linkml_meta": {"alias": "mappings", "domain_of": ["Node"]}},
    )
    external_links: Optional[List[ExpandedCurie]] = Field(
        None,
        description="""ExpandedCurie with id and url for xrefs""",
        json_schema_extra={"linkml_meta": {"alias": "external_links", "domain_of": ["Node"]}},
    )
    provided_by_link: Optional[ExpandedCurie] = Field(
        None,
        description="""A link to the docs for the knowledge source that provided the node/edge.""",
        json_schema_extra={"linkml_meta": {"alias": "provided_by_link", "domain_of": ["Association", "Node"]}},
    )
    association_counts: List[AssociationCount] = Field(
        ..., json_schema_extra={"linkml_meta": {"alias": "association_counts", "domain_of": ["Node"]}}
    )
    node_hierarchy: Optional[NodeHierarchy] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "node_hierarchy", "domain_of": ["Node"]}}
    )
    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    category: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )
    name: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "name", "domain_of": ["Entity", "MultiEntityAssociationResults"]}},
    )
    full_name: Optional[str] = Field(
        None,
        description="""The long form name of an entity""",
        json_schema_extra={"linkml_meta": {"alias": "full_name", "domain_of": ["Entity"]}},
    )
    deprecated: Optional[bool] = Field(
        None,
        description="""A boolean flag indicating that an entity is no longer considered current or valid.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "deprecated",
                "domain_of": ["Entity"],
                "exact_mappings": ["oboInOwl:ObsoleteClass"],
            }
        },
    )
    description: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "description", "domain_of": ["Entity"]}}
    )
    xref: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "xref", "domain_of": ["Entity"]}}
    )
    provided_by: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "provided_by", "domain_of": ["Association", "Entity"]}}
    )
    symbol: Optional[str] = Field(None, json_schema_extra={"linkml_meta": {"alias": "symbol", "domain_of": ["Entity"]}})
    synonym: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "synonym", "domain_of": ["Entity"]}}
    )
    uri: Optional[str] = Field(
        None,
        description="""The URI of the entity""",
        json_schema_extra={"linkml_meta": {"alias": "uri", "domain_of": ["Entity"]}},
    )
    iri: Optional[str] = Field(None, json_schema_extra={"linkml_meta": {"alias": "iri", "domain_of": ["Entity"]}})
    namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix portion of this entity's identifier""",
        json_schema_extra={"linkml_meta": {"alias": "namespace", "domain_of": ["Entity"]}},
    )
    has_phenotype: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype identifiers that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype", "domain_of": ["Entity"]}},
    )
    has_phenotype_label: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype labels that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_label", "domain_of": ["Entity"]}},
    )
    has_phenotype_closure: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype identifiers that are known to be associated with this entity expanded to include all ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_closure", "domain_of": ["Entity"]}},
    )
    has_phenotype_closure_label: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype labels that are known to be associated with this entity expanded to include all ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_closure_label", "domain_of": ["Entity"]}},
    )
    has_phenotype_count: Optional[int] = Field(
        None,
        description="""A count of the number of phenotypes that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_count", "domain_of": ["Entity"]}},
    )


class NodeHierarchy(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    super_classes: List[Entity] = Field(
        ..., json_schema_extra={"linkml_meta": {"alias": "super_classes", "domain_of": ["NodeHierarchy"]}}
    )
    sub_classes: List[Entity] = Field(
        ..., json_schema_extra={"linkml_meta": {"alias": "sub_classes", "domain_of": ["NodeHierarchy"]}}
    )


class Release(ConfiguredBaseModel):
    """
    A class to hold information about a release of the Monarch KG
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    version: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "version", "domain_of": ["Release"], "id_prefixes": ["string"]}},
    )
    url: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "url", "domain_of": ["ExpandedCurie", "Release"]}}
    )
    kg: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "kg", "domain_of": ["Release"], "id_prefixes": ["string"]}}
    )
    sqlite: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "sqlite", "domain_of": ["Release"], "id_prefixes": ["string"]}},
    )
    solr: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "solr", "domain_of": ["Release"], "id_prefixes": ["string"]}}
    )
    neo4j: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "neo4j", "domain_of": ["Release"], "id_prefixes": ["string"]}}
    )
    metadata: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "metadata", "domain_of": ["Release"], "id_prefixes": ["string"]}},
    )
    graph_stats: Optional[str] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "graph_stats", "domain_of": ["Release"], "id_prefixes": ["string"]}
        },
    )
    qc_report: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "qc_report", "domain_of": ["Release"], "id_prefixes": ["string"]}},
    )


class Results(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {"abstract": True, "from_schema": "https://w3id.org/monarch/monarch-py"}
    )

    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class AssociationResults(Results):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "Association"}},
        }
    )

    items: List[Association] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_fields: Optional[List[FacetField]] = Field(
        None,
        description="""Collection of facet field responses with the field values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_fields",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_queries: Optional[List[FacetValue]] = Field(
        None,
        description="""Collection of facet query responses with the query string values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_queries",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class CompactAssociationResults(Results):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "CompactAssociation"}},
        }
    )

    items: List[CompactAssociation] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_fields: Optional[List[FacetField]] = Field(
        None,
        description="""Collection of facet field responses with the field values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_fields",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_queries: Optional[List[FacetValue]] = Field(
        None,
        description="""Collection of facet query responses with the query string values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_queries",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class AssociationTableResults(Results):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "DirectionalAssociation"}},
        }
    )

    items: List[DirectionalAssociation] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_fields: Optional[List[FacetField]] = Field(
        None,
        description="""Collection of facet field responses with the field values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_fields",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_queries: Optional[List[FacetValue]] = Field(
        None,
        description="""Collection of facet query responses with the query string values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_queries",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class CategoryGroupedAssociationResults(Results):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "Association"}},
        }
    )

    counterpart_category: Optional[str] = Field(
        None,
        description="""The category of the counterpart entity in a given association,  eg. the category of the entity that is not the subject""",
        json_schema_extra={
            "linkml_meta": {"alias": "counterpart_category", "domain_of": ["CategoryGroupedAssociationResults"]}
        },
    )
    items: List[Association] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class EntityResults(Results):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "Entity"}},
        }
    )

    items: List[Entity] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class MappingResults(Results):
    """
    SSSOM Mappings returned as a results collection
    """

    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "Mapping"}},
        }
    )

    items: List[Mapping] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class MultiEntityAssociationResults(Results):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    name: Optional[str] = Field(
        None,
        json_schema_extra={"linkml_meta": {"alias": "name", "domain_of": ["Entity", "MultiEntityAssociationResults"]}},
    )
    associated_categories: List[CategoryGroupedAssociationResults] = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {"alias": "associated_categories", "domain_of": ["MultiEntityAssociationResults"]}
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class SearchResult(Entity):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {
                "category": {"name": "category", "required": True},
                "name": {"name": "name", "required": True},
            },
        }
    )

    score: Optional[float] = Field(
        None,
        json_schema_extra={
            "linkml_meta": {"alias": "score", "domain_of": ["BestMatch", "SemsimSearchResult", "SearchResult"]}
        },
    )
    id: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "id",
                "domain_of": [
                    "TermInfo",
                    "Association",
                    "ExpandedCurie",
                    "Entity",
                    "HistoPheno",
                    "HistoBin",
                    "Mapping",
                    "MultiEntityAssociationResults",
                ],
            }
        },
    )
    category: str = Field(
        ...,
        json_schema_extra={
            "linkml_meta": {
                "alias": "category",
                "domain_of": [
                    "Association",
                    "AssociationCount",
                    "CompactAssociation",
                    "AssociationTypeMapping",
                    "Entity",
                ],
            }
        },
    )
    name: str = Field(
        ...,
        json_schema_extra={"linkml_meta": {"alias": "name", "domain_of": ["Entity", "MultiEntityAssociationResults"]}},
    )
    full_name: Optional[str] = Field(
        None,
        description="""The long form name of an entity""",
        json_schema_extra={"linkml_meta": {"alias": "full_name", "domain_of": ["Entity"]}},
    )
    deprecated: Optional[bool] = Field(
        None,
        description="""A boolean flag indicating that an entity is no longer considered current or valid.""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "deprecated",
                "domain_of": ["Entity"],
                "exact_mappings": ["oboInOwl:ObsoleteClass"],
            }
        },
    )
    description: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "description", "domain_of": ["Entity"]}}
    )
    xref: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "xref", "domain_of": ["Entity"]}}
    )
    provided_by: Optional[str] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "provided_by", "domain_of": ["Association", "Entity"]}}
    )
    in_taxon: Optional[str] = Field(
        None,
        description="""The biolink taxon that the entity is in the closure of.""",
        json_schema_extra={"linkml_meta": {"alias": "in_taxon", "domain_of": ["Entity", "Node"]}},
    )
    in_taxon_label: Optional[str] = Field(
        None,
        description="""The label of the biolink taxon that the entity is in the closure of.""",
        json_schema_extra={"linkml_meta": {"alias": "in_taxon_label", "domain_of": ["Entity", "Node"]}},
    )
    symbol: Optional[str] = Field(None, json_schema_extra={"linkml_meta": {"alias": "symbol", "domain_of": ["Entity"]}})
    synonym: Optional[List[str]] = Field(
        None, json_schema_extra={"linkml_meta": {"alias": "synonym", "domain_of": ["Entity"]}}
    )
    uri: Optional[str] = Field(
        None,
        description="""The URI of the entity""",
        json_schema_extra={"linkml_meta": {"alias": "uri", "domain_of": ["Entity"]}},
    )
    iri: Optional[str] = Field(None, json_schema_extra={"linkml_meta": {"alias": "iri", "domain_of": ["Entity"]}})
    namespace: Optional[str] = Field(
        None,
        description="""The namespace/prefix portion of this entity's identifier""",
        json_schema_extra={"linkml_meta": {"alias": "namespace", "domain_of": ["Entity"]}},
    )
    has_phenotype: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype identifiers that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype", "domain_of": ["Entity"]}},
    )
    has_phenotype_label: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype labels that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_label", "domain_of": ["Entity"]}},
    )
    has_phenotype_closure: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype identifiers that are known to be associated with this entity expanded to include all ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_closure", "domain_of": ["Entity"]}},
    )
    has_phenotype_closure_label: Optional[List[str]] = Field(
        None,
        description="""A list of phenotype labels that are known to be associated with this entity expanded to include all ancestors""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_closure_label", "domain_of": ["Entity"]}},
    )
    has_phenotype_count: Optional[int] = Field(
        None,
        description="""A count of the number of phenotypes that are known to be associated with this entity""",
        json_schema_extra={"linkml_meta": {"alias": "has_phenotype_count", "domain_of": ["Entity"]}},
    )


class SearchResults(Results):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta(
        {
            "from_schema": "https://w3id.org/monarch/monarch-py",
            "slot_usage": {"items": {"name": "items", "range": "SearchResult"}},
        }
    )

    items: List[SearchResult] = Field(
        ...,
        description="""A collection of items, with the type to be overriden by slot_usage""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "items",
                "domain_of": [
                    "AssociationCountList",
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "CategoryGroupedAssociationResults",
                    "EntityResults",
                    "HistoPheno",
                    "MappingResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_fields: Optional[List[FacetField]] = Field(
        None,
        description="""Collection of facet field responses with the field values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_fields",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    facet_queries: Optional[List[FacetValue]] = Field(
        None,
        description="""Collection of facet query responses with the query string values and counts""",
        json_schema_extra={
            "linkml_meta": {
                "alias": "facet_queries",
                "domain_of": [
                    "AssociationResults",
                    "CompactAssociationResults",
                    "AssociationTableResults",
                    "SearchResults",
                ],
            }
        },
    )
    limit: int = Field(
        ...,
        description="""number of items to return in a response""",
        json_schema_extra={"linkml_meta": {"alias": "limit", "domain_of": ["Results"]}},
    )
    offset: int = Field(
        ...,
        description="""offset into the total number of items""",
        json_schema_extra={"linkml_meta": {"alias": "offset", "domain_of": ["Results"]}},
    )
    total: int = Field(
        ...,
        description="""total number of items matching a query""",
        json_schema_extra={"linkml_meta": {"alias": "total", "domain_of": ["Results"]}},
    )


class TextAnnotationResult(ConfiguredBaseModel):
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({"from_schema": "https://w3id.org/monarch/monarch-py"})

    text: Optional[str] = Field(
        None,
        description="""text without tokens""",
        json_schema_extra={"linkml_meta": {"alias": "text", "domain_of": ["TextAnnotationResult"]}},
    )
    tokens: Optional[List[Entity]] = Field(
        None,
        description="""A collection of entities or concepts""",
        json_schema_extra={"linkml_meta": {"alias": "tokens", "domain_of": ["TextAnnotationResult"]}},
    )
    start: Optional[int] = Field(
        None,
        description="""start position of the annotation""",
        json_schema_extra={"linkml_meta": {"alias": "start", "domain_of": ["TextAnnotationResult"]}},
    )
    end: Optional[int] = Field(
        None,
        description="""end position of the annotation""",
        json_schema_extra={"linkml_meta": {"alias": "end", "domain_of": ["TextAnnotationResult"]}},
    )


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
PairwiseSimilarity.model_rebuild()
TermPairwiseSimilarity.model_rebuild()
TermSetPairwiseSimilarity.model_rebuild()
TermInfo.model_rebuild()
BestMatch.model_rebuild()
SemsimSearchResult.model_rebuild()
Association.model_rebuild()
AssociationCountList.model_rebuild()
CompactAssociation.model_rebuild()
AssociationTypeMapping.model_rebuild()
DirectionalAssociation.model_rebuild()
ExpandedCurie.model_rebuild()
Entity.model_rebuild()
FacetValue.model_rebuild()
AssociationCount.model_rebuild()
FacetField.model_rebuild()
HistoPheno.model_rebuild()
HistoBin.model_rebuild()
Mapping.model_rebuild()
Node.model_rebuild()
NodeHierarchy.model_rebuild()
Release.model_rebuild()
Results.model_rebuild()
AssociationResults.model_rebuild()
CompactAssociationResults.model_rebuild()
AssociationTableResults.model_rebuild()
CategoryGroupedAssociationResults.model_rebuild()
EntityResults.model_rebuild()
MappingResults.model_rebuild()
MultiEntityAssociationResults.model_rebuild()
SearchResult.model_rebuild()
SearchResults.model_rebuild()
TextAnnotationResult.model_rebuild()



import sys
from typing import List

import bs4
import requests
from rich.console import Console

from monarch_py.datamodels.model import ExpandedCurie, Release
from monarch_py.service.curie_service import converter

KG_URL = "https://data.monarchinitiative.org/monarch-kg"
KG_DEV_URL = "https://data.monarchinitiative.org/monarch-kg-dev"
SOLR_DATA_URL = f"{KG_DEV_URL}/latest/solr.tar.gz"
SQL_DATA_URL = f"{KG_DEV_URL}/latest/monarch-kg.db.gz"


console = Console(
    color_system="truecolor",
    stderr=True,
    style="pink1",
)


def strip_json(doc: dict, *fields_to_remove: str):
    for field in fields_to_remove:
        try:
            del doc[field]
        except KeyError:
            pass
    return doc


def escape(value: str) -> str:
    return value.replace(":", r"\:")


def dict_factory(cursor, row):
    """Converts a sqlite3 row to a dictionary."""
    fields = [column[0] for column in cursor.description]
    return {key: value for key, value in zip(fields, row)}


def compare_dicts(dict1, dict2):
    """Compare two dictionaries"""
    return all([k in dict2 for k in dict1]) and all([dict1[k] == dict2[k] for k in dict2])


def dict_diff(d1, d2) -> dict:
    """Return the difference between two dictionaries"""
    difference = {}
    for k in d1.keys():
        if d1[k] != d2[k]:
            difference[f"Dict-1-{k}"] = d1[k]
            difference[f"Dict-2-{k}"] = d2[k]
    return difference


def set_log_level(log_level: str):
    """Sets the log level for the application."""
    import loguru

    loguru.logger.remove()
    loguru.logger.add(sys.stderr, level=log_level)


### URL fetching methods ###


def get_links_for_field(field: List[str]) -> List[ExpandedCurie]:
    # TODO should be able to remove curie.replace("PMID", "PUBMED")) since the converter should handle prefix synonyms
    expanded_curies = [ExpandedCurie(id=curie, url=get_link_for_curie(curie)) for curie in field if ":" in curie]
    return expanded_curies


def get_link_for_curie(curie: str) -> ExpandedCurie:
    url = converter.expand(curie.replace("PMID", "PUBMED"))
    if curie.startswith("GARD:"):
        url += "/index"
    return url


def get_provided_by_link(provided_by: str) -> ExpandedCurie:
    """Returns a link to the provided_by resource."""
    base_url = "https://monarch-initiative.github.io/monarch-ingest/Sources"
    pb = provided_by.replace("_nodes", "").replace("_edges", "").split("_")
    slug = f"{pb[0]}/#{'_'.join(pb[1:])}"
    return ExpandedCurie(
        id=provided_by.replace("_nodes", "").replace("_edges", "") if provided_by else None, url=f"{base_url}/{slug}"
    )


### Release info methods ###


def get_release_versions(dev: bool = False, limit: int = 0, print_info: bool = False):
    url = f"{KG_DEV_URL if dev else KG_URL}/index.html"
    res = requests.get(url)
    res.raise_for_status()
    soup = bs4.BeautifulSoup(res.text, "html.parser")
    directories = soup.select("h5 + ul > li > a")
    releases = [
        {"version": directory.text, "url": directory["href"]}
        for directory in directories
        if directory.text not in ["..", "kgx"]
        # if directory.text != ".."]
    ]
    releases.reverse()
    if limit:
        releases = releases[:limit]
    if print_info:
        for release in releases:
            console.print(f"Version: {release['version']:11s}{' ':—<3} URL: {release['url']}")
    return releases


def get_release_metadata(release: str, dev: bool = False):
    release_url = f"{KG_DEV_URL if dev else KG_URL}/{release}"
    release_info = Release(
        version=release,
        url=f"{release_url}/index.html",
        kg=f"{release_url}/monarch-kg.tar.gz",
        sqlite=f"{release_url}/monarch-kg.db.gz",
        solr=f"{release_url}/solr.tar.gz",
        neo4j=f"{release_url}/monarch-kg.neo4j.dump",
        metadata=f"{release_url}/metadata.yaml",
        graph_stats=f"{release_url}/merged_graph_stats.yaml",
        qc_report=f"{release_url}/qc_report.yaml",
    )
    return release_info


def print_release_info(release_info: Release):
    for key, value in release_info.model_dump().items():
        console.print(f"{key+' ':—<12} {value}")



from abc import ABC, abstractmethod

from monarch_py.datamodels.model import Entity


class EntityInterface(ABC):
    """Abstract interface for entities in the Monarch KG"""

    @abstractmethod
    def get_entity(self, id: str, get_association_counts: bool = False, get_hierarchy: bool = False) -> Entity:
        """Retrieve a specific entity by exact ID match, with optional extras

        Args:
            id (str): id of the entity to search for.
            get_association_counts (bool, optional): Whether to get a count of associations for the entity. Defaults to False.
            get_hierarchy (bool, optional): Whether to get the entity's heirarchy in the data model. Defaults to False.

        Raises:
            NotImplementedError: Use a specific implementation (see the documentation for a list of implementations)

        Returns:
            Entity: Dataclass representing results of an entity search.
        """
        raise NotImplementedError



import importlib
import importlib.util
from pathlib import Path
from typing import Annotated, Optional

import typer

from monarch_py import solr_cli, sql_cli
from monarch_py.api.config import semsimian
from monarch_py.api.additional_models import SemsimMetric
from monarch_py.utils.solr_cli_utils import check_for_docker
from monarch_py.utils.utils import (
    set_log_level,
    get_release_metadata,
    get_release_versions,
)
from monarch_py.utils.format_utils import format_output
from monarch_py.utils import cli_fields as fields


app = typer.Typer()
app.add_typer(solr_cli.solr_app, name="solr")
app.add_typer(sql_cli.sql_app, name="sql")


@app.callback(invoke_without_command=True)
def callback(
    ctx: typer.Context,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-v",
            help="Show the currently installed version",
            is_eager=True,
        ),
    ] = None,
    quiet: Annotated[
        bool,
        typer.Option(
            "--quiet",
            "-q",
            help="Set log level to warning",
        ),
    ] = False,
    debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            "-d",
            help="Set log level to debug",
        ),
    ] = False,
    # verbose: Annotated[int, typer.Option("--verbose", "-v", count=True)] = 0,
):
    if version and ctx.invoked_subcommand is None:
        from monarch_py import __version__

        typer.echo(f"monarch_py version: {__version__}")
        raise typer.Exit()
    if ctx.invoked_subcommand is None:
        typer.secho(
            f"\n\tNo command specified\n\tTry `monarch --help` for more information.\n",
            fg=typer.colors.YELLOW,
        )
        raise typer.Exit()
    if ctx.invoked_subcommand != "compare":
        check_for_docker()
    set_log_level(log_level="DEBUG" if debug else "WARNING" if quiet else "INFO")
    return


@app.command("test")
def test():
    """Test the CLI"""
    typer.secho("\n\tTesting monarch_py CLI\n", fg=typer.colors.GREEN)


@app.command("schema")
def schema():
    """
    Print the linkml schema for the data model
    """
    schema_name = "model"
    import_path = f"monarch_py.datamodels.{schema_name}"
    schema_module_spec = importlib.util.find_spec(f"monarch_py.datamodels.{schema_name}")
    if schema_module_spec is None or schema_module_spec.origin is None:
        print(f"No python module found at {import_path}")
        raise typer.Exit(code=1)
    schema_dir = Path(schema_module_spec.origin).parent
    schema_path = schema_dir / f"{schema_name}.yaml"
    with open(schema_path, "r") as schema_file:
        print(schema_file.read())
    raise typer.Exit()


### "Aliases" for Solr CLI ###


app.command("entity")(solr_cli.entity)
app.command("associations")(solr_cli.associations)
app.command("multi-entity-associations")(solr_cli.multi_entity_associations)
app.command("search")(solr_cli.search)
app.command("autocomplete")(solr_cli.autocomplete)
app.command("histopheno")(solr_cli.histopheno)
app.command("association-counts")(solr_cli.association_counts)
app.command("association-table")(solr_cli.association_table)
app.command("mappings")(solr_cli.mappings)


### CLI Commands for Semsimian ###


@app.command("compare")
def compare(
    subjects: Annotated[
        str,
        typer.Argument(
            help="Comma separated list of subjects to compare",
        ),
    ],
    objects: Annotated[
        str,
        typer.Argument(
            help="Comma separated list of objects to compare",
        ),
    ],
    metric: Annotated[
        SemsimMetric,
        typer.Option(
            "--metric",
            "-m",
            help="The metric to use for comparison",
        ),
    ] = SemsimMetric.ANCESTOR_INFORMATION_CONTENT,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """Compare two sets of phenotypes using semantic similarity via SemSimian"""
    subjects_list = subjects.split(",")
    objects_list = objects.split(",")
    response = semsimian().compare(subjects_list, objects_list, metric)
    format_output(fmt, response, output)


### CLI Commands for Release Info ###


@app.command("releases")
def releases(
    dev: Annotated[
        bool,
        typer.Option(
            "--dev",
            help="Get dev releases of the KG (default is False)",
        ),
    ] = False,
    limit: fields.LimitOption = 0,
):
    """
    List all available releases of the Monarch Knowledge Graph
    """
    get_release_versions(dev, limit, print_info=True)


@app.command("release")
def release(
    release_ver: str = typer.Argument(help="The release version to get metadata for"),
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Retrieve metadata for a specific release
    """
    release_info = get_release_metadata(release_ver)
    format_output(fmt, release_info, output)


if __name__ == "__main__":
    app()



### make a singleton class that uses prefixmap and curies to expand curies

from prefixmaps import load_converter

__all__ = [
    "converter",
]

# this is a magic keyword that represents the "merged" context from Chris M's algorithm
# (https://github.com/linkml/prefixmaps/blob/main/src/prefixmaps/data/merged.csv)
converter = load_converter("merged")
converter.add_prefix("GARD", "https://rarediseases.info.nih.gov/diseases/")
converter.add_prefix("Orphanet", "https://www.orpha.net/en/disease/detail/", merge=True)
